from typing import Any, Awaitable, Callable

from xpresso.openapi.models import SecurityScheme


class SecurityBase:
    model: SecurityScheme
    scheme_name: str
    __call__: Callable[..., Awaitable[Any]]



import inspect
import typing
from dataclasses import dataclass, field

from pydantic.fields import (
    SHAPE_FROZENSET,
    SHAPE_ITERABLE,
    SHAPE_LIST,
    SHAPE_SEQUENCE,
    SHAPE_SET,
    SHAPE_TUPLE_ELLIPSIS,
    ModelField,
)
from pydantic.schema import get_flat_models_from_field

from xpresso._utils.schemas import openapi_schema_from_pydantic_field
from xpresso._utils.typing import is_sequence_like, model_field_from_param
from xpresso.binders._utils.examples import parse_examples
from xpresso.binders.api import (
    ModelNameMap,
    OpenAPIParameter,
    OpenAPIParameterMarker,
    Schemas,
)
from xpresso.openapi import models as openapi_models

Examples = typing.Optional[typing.Mapping[str, openapi_models.Example]]

SEQUENCE_LIKE_SHAPES = (
    SHAPE_LIST,
    SHAPE_SET,
    SHAPE_SEQUENCE,
    SHAPE_ITERABLE,
    SHAPE_TUPLE_ELLIPSIS,
    SHAPE_FROZENSET,
)


@dataclass(frozen=True)
class OpenAPIParameterBase(OpenAPIParameter):
    name: str
    in_: str
    required: bool
    style: str
    explode: bool
    description: typing.Optional[str]
    deprecated: typing.Optional[bool]
    include_in_schema: bool
    examples: Examples = field(compare=False)
    field: ModelField = field(compare=False)
    param_cls: typing.ClassVar[typing.Type[openapi_models.ConcreteParameter]]

    def get_openapi(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.ConcreteParameter:
        return self.param_cls(
            description=self.description or self.field.field_info.description,  # type: ignore[arg-type]
            required=None if self.required is False else True,
            deprecated=self.deprecated,
            style=self.style,  # type: ignore[arg-type]
            explode=self.explode,
            schema=openapi_schema_from_pydantic_field(
                self.field, model_name_map, schemas
            ),
            examples=self.examples,  # type: ignore[arg-type]
            name=self.name,
        )

    def get_models(self) -> typing.List[type]:
        return list(get_flat_models_from_field(self.field, known_models=set()))


@dataclass(frozen=True)
class OpenAPIParameterMarkerBase(OpenAPIParameterMarker):
    alias: typing.Optional[str]
    style: str
    explode: bool
    description: typing.Optional[str]
    deprecated: typing.Optional[bool]
    include_in_schema: bool
    examples: Examples
    in_: typing.ClassVar[str]
    cls: typing.ClassVar[typing.Type[OpenAPIParameterBase]]
    required: typing.ClassVar[bool] = False

    def register_parameter(self, param: inspect.Parameter) -> OpenAPIParameterBase:
        field = model_field_from_param(param)
        name = self.alias or field.alias
        if field.required is False and self.required:
            raise TypeError(
                f"{self.in_.title()} parameters MUST be required and MUST NOT have default values"
            )
        if self.required:
            required = True
        elif field.required and not is_sequence_like(field):
            required = True
        else:
            required = False
        return self.cls(
            name=name,
            in_=self.in_,
            required=required,
            field=field,
            style=self.style,
            explode=self.explode,
            description=self.description,
            deprecated=self.deprecated,
            include_in_schema=self.include_in_schema,
            examples=parse_examples(self.examples) if self.examples else None,
        )



import binascii
from base64 import b64decode
from typing import NamedTuple, Optional

from pydantic import BaseModel
from starlette.requests import Request
from starlette.status import HTTP_401_UNAUTHORIZED

from xpresso.binders._security.utils import get_authorization_scheme_param
from xpresso.binders.api import SecurityBase
from xpresso.exceptions import HTTPException
from xpresso.openapi.models import HTTPBase as HTTPBaseModel
from xpresso.openapi.models import HTTPBearer as HTTPBearerModel


class HTTPBasicCredentials(BaseModel):
    username: str
    password: str


class HTTPAuthorizationCredentials(BaseModel):
    scheme: str
    credentials: str


class HTTPBaseConfig(NamedTuple):
    scheme: str
    scheme_name: Optional[str] = None
    description: Optional[str] = None
    auto_error: bool = True


class HTTPBase(SecurityBase):
    def __init__(
        self,
        *,
        scheme: str,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        auto_error: bool = True,
    ):
        self.model = HTTPBaseModel(scheme=scheme, description=description)
        self.scheme_name = scheme_name or self.__class__.__name__
        self.auto_error = auto_error

    async def __call__(  # type: ignore  # for Pylance
        self, request: Request
    ) -> Optional[HTTPAuthorizationCredentials]:
        authorization: str = request.headers.get("Authorization")
        scheme, credentials = get_authorization_scheme_param(authorization)
        if not (authorization and scheme and credentials):
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED, detail="Not authenticated"
                )
            else:
                return None
        return HTTPAuthorizationCredentials(scheme=scheme, credentials=credentials)


class HTTPBasic(HTTPBase):
    def __init__(
        self,
        *,
        scheme_name: Optional[str] = None,
        realm: Optional[str] = None,
        description: Optional[str] = None,
        auto_error: bool = True,
    ):
        self.model = HTTPBaseModel(scheme="basic", description=description)
        self.scheme_name = scheme_name or self.__class__.__name__
        self.realm = realm
        self.auto_error = auto_error

    async def __call__(self, request: Request) -> Optional[HTTPBasicCredentials]:  # type: ignore
        authorization: str = request.headers.get("Authorization")
        scheme, param = get_authorization_scheme_param(authorization)
        if self.realm:
            unauthorized_headers = {"WWW-Authenticate": f'Basic realm="{self.realm}"'}
        else:
            unauthorized_headers = {"WWW-Authenticate": "Basic"}
        invalid_user_credentials_exc = HTTPException(
            status_code=HTTP_401_UNAUTHORIZED,
            detail="Invalid authentication credentials",
            headers=unauthorized_headers,
        )
        if not authorization or scheme.lower() != "basic":
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED,
                    detail="Not authenticated",
                    headers=unauthorized_headers,
                )
            else:
                return None
        try:
            data = b64decode(param).decode("ascii")
        except (ValueError, UnicodeDecodeError, binascii.Error):
            raise invalid_user_credentials_exc
        username, separator, password = data.partition(":")
        if not separator:
            raise invalid_user_credentials_exc
        return HTTPBasicCredentials(username=username, password=password)


class HTTPBearer(HTTPBase):
    def __init__(
        self,
        *,
        bearerFormat: Optional[str] = None,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        auto_error: bool = True,
    ):
        self.model = HTTPBearerModel(bearerFormat=bearerFormat, description=description)  # type: ignore  # for Pylance
        self.scheme_name = scheme_name or self.__class__.__name__
        self.auto_error = auto_error

    async def __call__(
        self, request: Request
    ) -> Optional[HTTPAuthorizationCredentials]:
        authorization: str = request.headers.get("Authorization")
        scheme, credentials = get_authorization_scheme_param(authorization)
        if not (authorization and scheme and credentials):
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED, detail="Not authenticated"
                )
            else:
                return None
        if scheme.lower() != "bearer":
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED,
                    detail="Invalid authentication credentials",
                )
            else:
                return None
        return HTTPAuthorizationCredentials(scheme=scheme, credentials=credentials)


class HTTPDigest(HTTPBase):
    def __init__(
        self,
        *,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        auto_error: bool = True,
    ):
        self.model = HTTPBaseModel(scheme="digest", description=description)
        self.scheme_name = scheme_name or self.__class__.__name__
        self.auto_error = auto_error

    async def __call__(
        self, request: Request
    ) -> Optional[HTTPAuthorizationCredentials]:
        authorization: str = request.headers.get("Authorization")
        scheme, credentials = get_authorization_scheme_param(authorization)
        if not (authorization and scheme and credentials):
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED, detail="Not authenticated"
                )
            else:
                return None
        if scheme.lower() != "digest":
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED,
                detail="Invalid authentication credentials",
            )
        return HTTPAuthorizationCredentials(scheme=scheme, credentials=credentials)



from xpresso.binders._parameters.openapi.base import (
    OpenAPIParameterBase,
    OpenAPIParameterMarkerBase,
)
from xpresso.openapi import models as openapi_models


class OpenAPICookieParameter(OpenAPIParameterBase):
    param_cls = openapi_models.Cookie


class OpenAPICookieParameterMarker(OpenAPIParameterMarkerBase):
    cls = OpenAPICookieParameter
    in_ = "cookie"



import inspect
import typing
from dataclasses import dataclass

from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import ModelField
from starlette.requests import HTTPConnection

from xpresso._utils.typing import model_field_from_param
from xpresso.binders.api import ParameterExtractor
from xpresso.exceptions import RequestValidationError, WebSocketValidationError
from xpresso.typing import Some


@dataclass(frozen=True)
class ParameterExtractorBase(ParameterExtractor):
    field: ModelField
    loc: typing.Tuple[str, ...]
    name: str
    in_: typing.ClassVar[str]

    async def validate(
        self,
        values: typing.Optional[Some[typing.Any]],
        connection: HTTPConnection,
    ) -> typing.Any:
        """Validate after parsing. Only used by the top-level body"""
        if values is None:
            if self.field.required is False:
                return self.field.get_default()
            else:
                if connection.scope["type"] == "websocket":
                    raise WebSocketValidationError(
                        [
                            ErrorWrapper(
                                ValueError(f"Missing required {self.in_} parameter"),
                                loc=self.loc,
                            )
                        ]
                    )
                raise RequestValidationError(
                    [
                        ErrorWrapper(
                            ValueError(f"Missing required {self.in_} parameter"),
                            loc=self.loc,
                        )
                    ]
                )
        val, errs = self.field.validate(values.value, {}, loc=self.loc)
        if errs:
            if isinstance(errs, ErrorWrapper):
                errs = [errs]
            errs = typing.cast(typing.List[ErrorWrapper], errs)
            if connection.scope["type"] == "websocket":
                raise WebSocketValidationError(errs)
            raise RequestValidationError(errs)
        return val


def get_basic_param_info(
    param: inspect.Parameter, alias: typing.Optional[str], in_: str
) -> typing.Tuple[ModelField, str, typing.Tuple[str, ...]]:
    field = model_field_from_param(param)
    name = alias or field.alias
    loc = (in_, name)
    return (field, name, loc)



import inspect
import typing

from di import BaseContainer, Dependant, SyncExecutor
from di.api.dependencies import CacheKey, DependantBase

import xpresso.openapi.models as openapi_models
from xpresso.binders.api import (
    BodyExtractor,
    BodyExtractorMarker,
    OpenAPIBody,
    OpenAPIBodyMarker,
    OpenAPIParameter,
    OpenAPIParameterMarker,
    ParameterExtractor,
    ParameterExtractorMarker,
    SecurityBase,
)
from xpresso.exceptions import XpressoError


class ParameterBinderMarker(Dependant[typing.Any]):
    def __init__(
        self,
        *,
        in_: str,
        extractor_marker: ParameterExtractorMarker,
        openapi_marker: OpenAPIParameterMarker,
    ) -> None:
        super().__init__(call=None, scope="connection")
        self.in_ = in_
        self.extractor_marker = extractor_marker
        self.openapi_marker = openapi_marker

    def register_parameter(self, param: inspect.Parameter) -> DependantBase[typing.Any]:
        return ParameterBinder(
            in_=self.in_,
            openapi=self.openapi_marker.register_parameter(param),
            extractor=self.extractor_marker.register_parameter(param),
        )


class ParameterBinder(Dependant[typing.Any]):
    __slots__ = (
        "in_",
        "alias",
        "openapi",
    )

    def __init__(
        self,
        in_: str,
        openapi: OpenAPIParameter,
        extractor: ParameterExtractor,
    ):
        self.in_ = in_
        self.openapi = openapi
        super().__init__(call=extractor.extract, scope="connection")

    @property
    def cache_key(self) -> CacheKey:
        # As per the spec, parameters are identified
        # by name and location
        return (self.__class__, self.openapi.name, self.in_)


class BodyBinder(Dependant[typing.Any]):
    __slots__ = ("openapi", "extractor")

    def __init__(
        self,
        *,
        openapi: OpenAPIBody,
        extractor: BodyExtractor,
    ) -> None:
        super().__init__(call=extractor.extract_from_request, scope="connection")
        self.openapi = openapi
        self.extractor = extractor


class BodyBinderMarker(Dependant[typing.Any]):
    def __init__(
        self,
        *,
        extractor_marker: BodyExtractorMarker,
        openapi_marker: OpenAPIBodyMarker,
    ) -> None:
        super().__init__(call=None, scope="connection")
        self.extractor_marker = extractor_marker
        self.openapi_marker = openapi_marker

    def register_parameter(self, param: inspect.Parameter) -> BodyBinder:
        return BodyBinder(
            openapi=self.openapi_marker.register_parameter(param),
            extractor=self.extractor_marker.register_parameter(param),
        )


class SecurityBinder(Dependant[typing.Any]):
    _model: typing.Optional[openapi_models.SecurityScheme]

    def __init__(
        self,
        *,
        scheme_name: typing.Optional[str] = None,
        model: typing.Union[SecurityBase, typing.Type[SecurityBase]],
        scopes: typing.Optional[typing.Sequence[str]],
    ) -> None:
        self.model = model
        self.scopes = frozenset(scopes or [])
        if isinstance(model, SecurityBase):

            self._model = model.model
            self._scheme_name = scheme_name or model.__class__.__name__

            def get_model() -> SecurityBase:
                assert isinstance(model, SecurityBase)
                return model

            extractor = model.__class__.__call__
            self._model_dependant = Dependant(get_model, scope="app")
            super().__init__(
                call=extractor,
                overrides={"self": self._model_dependant},
                scope="connection",
            )
        else:
            self._model = None
            self._scheme_name = scheme_name or model.__name__
            self._model_dependant = Dependant(model, scope="app")
            super().__init__(
                call=model.__call__,
                overrides={"self": self._model_dependant},
                scope="connection",
            )

    @property
    def scheme_name(self) -> str:
        return self._scheme_name

    def construct_model(
        self, container: BaseContainer
    ) -> openapi_models.SecurityScheme:
        if self._model is not None:
            return self._model
        if "app" not in container.current_scopes:
            raise XpressoError(
                "Defered security models require a lifespan event to be triggered to be able to generate OpenAPI schemas"
                "\nIf you are using TestClient, use it as a context manager:"
                "\n  with TestClient(app) as client:  ..."
            )
        solved = container.solve(self._model_dependant)
        model = container.execute_sync(solved, executor=SyncExecutor())
        return model.model

    def register_parameter(self, param: inspect.Parameter) -> DependantBase[typing.Any]:
        return self

    @property
    def cache_key(self) -> CacheKey:
        return (self.__class__, self.model)



from xpresso.binders._parameters.openapi.base import (
    OpenAPIParameterBase,
    OpenAPIParameterMarkerBase,
)
from xpresso.openapi import models as openapi_models


class OpenAPIHeaderParameter(OpenAPIParameterBase):
    param_cls = openapi_models.Header


class OpenAPIHeaderParameterMarker(OpenAPIParameterMarkerBase):
    cls = OpenAPIHeaderParameter
    in_ = "header"



from xpresso.binders._parameters.openapi.base import (
    OpenAPIParameterBase,
    OpenAPIParameterMarkerBase,
)
from xpresso.openapi import models as openapi_models


class OpenAPIQueryParameter(OpenAPIParameterBase):
    param_cls = openapi_models.Query


class OpenAPIQueryParameterMarker(OpenAPIParameterMarkerBase):
    cls = OpenAPIQueryParameter
    in_ = "header"



from __future__ import annotations

from typing import Optional, Sequence, Union

from di.api.providers import DependencyProviderType

from xpresso.binders.api import SecurityBase
from xpresso.dependencies.models import Dependant


class Security(Dependant):
    _dependency: Union[DependencyProviderType[SecurityBase], SecurityBase]

    def __init__(
        self,
        dependency: Union[DependencyProviderType[SecurityBase], SecurityBase],
        *,
        scopes: Optional[Sequence[str]] = None,
    ):
        self._dependency = dependency
        if isinstance(dependency, SecurityBase):

            async def call() -> SecurityBase:
                return dependency  # type: ignore[return-value]

            super().__init__(call, scope="app")
        else:
            super().__init__(dependency, scope="app")
        self.scopes = frozenset(scopes or [])

    @property
    def cache_key(self) -> Union[DependencyProviderType[SecurityBase], SecurityBase]:
        return self._dependency



import inspect
import sys
from collections import ChainMap
from typing import Any, Callable, Dict, Iterable, List, Optional, TypeVar, Union

if sys.version_info < (3, 9):
    from typing_extensions import get_args, get_origin, get_type_hints
else:
    from typing import get_origin, get_args, get_type_hints

from pydantic import BaseConfig
from pydantic.fields import ModelField
from pydantic.schema import field_schema, get_flat_models_from_field
from pydantic.typing import NoneType
from starlette.responses import Response

from xpresso._utils.typing import get_model_name_map
from xpresso.openapi import models
from xpresso.openapi.constants import REF_PREFIX
from xpresso.responses import JsonResponseSpec, ResponseSpec

T = TypeVar("T")

ModelNameMap = Dict[type, str]


def _get_response_models_from_return_hint(
    return_hint: type,
) -> List[ResponseSpec]:
    res: Dict[str, ResponseSpec] = {}
    description = ""
    if get_origin(return_hint) is Union:
        tps: List[Any] = []
        for tp in get_args(return_hint):
            if inspect.isclass(tp) and issubclass(tp, Response):
                if tp.media_type is not None:
                    res[tp.media_type] = ResponseSpec(
                        description=description,
                        media_type=tp.media_type,
                    )
            else:
                if inspect.isclass(tp) and issubclass(tp, NoneType):
                    return [ResponseSpec(description=description)]
                tps.append(tp)
        if tps:
            res["application/json"] = JsonResponseSpec(
                description=description,
                model=Union[tuple(tps)],
            )
    else:
        tp = return_hint
        if inspect.isclass(tp) and issubclass(tp, Response):
            if tp.media_type is not None:
                res[tp.media_type] = ResponseSpec(
                    description=description,
                    media_type=tp.media_type,
                )
        else:
            if inspect.isclass(tp) and issubclass(tp, NoneType):
                return [ResponseSpec(description=description)]
            res["application/json"] = JsonResponseSpec(
                description=description,
                model=tp,
            )
    return list(res.values()) or [ResponseSpec(description=description)]


def get_response_specs_from_return_type_hints(
    endpoint: Callable[..., Any]
) -> List[ResponseSpec]:
    hints = get_type_hints(endpoint, include_extras=True)
    if "return" not in hints:
        return []
    return _get_response_models_from_return_hint(hints["return"])


def _get_response_schema(
    type_: type, model_name_map: ModelNameMap, schemas: Dict[str, Any]
) -> Dict[str, Any]:
    field = ModelField.infer(
        name="response",
        value=...,
        annotation=type_,
        class_validators={},
        config=BaseConfig,
    )
    model_name_map.update(
        get_model_name_map(
            get_flat_models_from_field(
                field,
                model_name_map.keys(),  # type: ignore[arg-type]
            )
        )
    )
    schema, refs, _ = field_schema(
        field, by_alias=True, ref_prefix=REF_PREFIX, model_name_map=model_name_map
    )
    schemas.update(refs)
    return schema


def get_response(
    spec: ResponseSpec,
    model_name_map: ModelNameMap,
    schemas: Dict[str, Any],
) -> models.Response:
    if spec.model is not None and spec.media_type is not None:
        schema = _get_response_schema(spec.model, model_name_map, schemas)
        examples: Optional[Dict[str, models.Example]]
        if spec.examples:
            examples = {
                n: ex if isinstance(ex, models.Example) else models.Example(value=ex)
                for n, ex in spec.examples.items()
            }
        else:
            examples = None
        content = {spec.media_type: {"schema": schema, "examples": examples}}
        model = models.Response(
            description=spec.description, headers=spec.headers, content=content
        )
    else:
        model = models.Response(
            description=spec.description,
            headers=spec.headers,
            content=None,
        )

    return model


def merge_response_models(
    mods: Iterable[models.Response], default_description: Optional[str] = None
) -> models.Response:
    mods = list(mods)
    if len(mods) == 1:
        mod = next(iter(mods))
        return models.Response(
            description=mod.description or default_description or "",
            headers=mod.headers,
            content=mod.content,
        )
    descriptions: List[str] = []
    for m in mods:
        if m.content and m.description:
            descriptions.append(f"- {next(iter(m.content.keys()))}: {m.description}")
    description = "\n".join(descriptions)
    if default_description:
        description = description or default_description
    headers = dict(ChainMap(*((m.headers or {}) for m in mods))) or None
    content = dict(ChainMap(*((m.content or {}) for m in mods))) or None
    return models.Response(description=description, headers=headers, content=content)



from collections import Counter
from typing import Dict, Mapping, Tuple

from di import BaseContainer

from xpresso.binders import dependants as binder_dependants
from xpresso.openapi import models
from xpresso.routing.operation import Operation
from xpresso.routing.pathitem import Path

SecurityModels = Mapping[binder_dependants.SecurityBinder, models.SecurityScheme]


def get_security_models(
    routes: Mapping[str, Tuple[Path, Mapping[str, Operation]]], container: BaseContainer
) -> Mapping[binder_dependants.SecurityBinder, models.SecurityScheme]:
    res: Dict[binder_dependants.SecurityBinder, models.SecurityScheme] = {}
    for _, operations in routes.values():
        for operation in operations.values():
            dependant = operation.dependant
            flat_dependencies = dependant.get_flat_subdependants()
            for dep in flat_dependencies:
                if isinstance(
                    dep,
                    binder_dependants.SecurityBinder,
                ):
                    res[dep] = dep.construct_model(container)
    return res


def get_security_scheme_name_map(
    models: Mapping[binder_dependants.SecurityBinder, models.SecurityScheme]
) -> Mapping[models.SecurityScheme, str]:
    scheme_names = {scheme: binder.scheme_name for binder, scheme in models.items()}
    name_counter = Counter(scheme_names.values())
    schemes_with_duplicate_names = [
        scheme for scheme in scheme_names if name_counter[scheme_names[scheme]] > 1
    ]
    # sort in reverse order so that APIKey(name="key1") gets assigned the deduped name "APIKey_1"
    for scheme in sorted(
        schemes_with_duplicate_names,
        key=lambda scheme: tuple(scheme.dict().items()),
        reverse=True,
    ):
        name = scheme_names[scheme]
        scheme_names[scheme] += f"_{name_counter[name]}"
        name_counter[name] -= 1
    return scheme_names



import inspect
import sys
from typing import Any, Awaitable, Callable, Dict, Iterable, List, Optional, Union

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

from starlette.datastructures import FormData, UploadFile
from starlette.requests import HTTPConnection, Request

from xpresso.openapi import models
from xpresso.typing import Some

Model = type
ModelNameMap = Dict[Model, str]
Schemas = Dict[str, Union[models.Schema, models.Reference]]


class BodyExtractor(Protocol):
    async def extract_from_request(self, request: Request) -> Any:
        """Extract from top level request"""
        raise NotImplementedError

    async def extract_from_form(
        self,
        form: FormData,
        *,
        loc: Iterable[Union[str, int]],
    ) -> Optional[Some[Any]]:
        """Extract from a form"""
        raise NotImplementedError

    async def extract_from_field(
        self,
        field: Union[str, UploadFile],
        *,
        loc: Iterable[Union[str, int]],
    ) -> Any:
        """Extract from a form field"""
        raise NotImplementedError

    def matches_media_type(self, media_type: Optional[str]) -> bool:
        """Check if this extractor can extract the given media type"""
        raise NotImplementedError


class BodyExtractorMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> BodyExtractor:
        """Hook to pull information from the Python parameter/field this body is attached to"""
        raise NotImplementedError


class ParameterExtractor(Protocol):
    def extract(
        self,
        connection: HTTPConnection,
    ) -> Any:
        raise NotImplementedError


class ParameterExtractorMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> ParameterExtractor:
        """Hook to pull information from the Python parameter/field this body is attached to"""
        raise NotImplementedError


class OpenAPIBody(Protocol):
    include_in_schema: bool

    def get_models(self) -> List[type]:
        """The type representing this parameter.

        This is used as the key to deduplicate references and schema names.
        """
        raise NotImplementedError

    def get_openapi(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.RequestBody:
        raise NotImplementedError

    def get_schema(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.Schema:
        raise NotImplementedError

    def get_media_type_string(self) -> str:
        raise NotImplementedError

    def get_openapi_media_type(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.MediaType:
        raise NotImplementedError

    def get_encoding(self) -> Optional[models.Encoding]:
        raise NotImplementedError

    def get_field_name(self) -> str:
        raise NotImplementedError


class OpenAPIBodyMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> OpenAPIBody:
        raise NotImplementedError


class OpenAPIParameter(Protocol):
    in_: str
    name: str
    include_in_schema: bool

    def get_openapi(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.ConcreteParameter:
        raise NotImplementedError

    def get_models(self) -> List[type]:
        """Used as the key to deduplicate references and schema names"""
        raise NotImplementedError


class OpenAPIParameterMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> OpenAPIParameter:
        raise NotImplementedError


class SecurityBase:
    model: models.SecurityScheme
    scheme_name: str
    __call__: Callable[..., Awaitable[Any]]



import sys
from typing import Any

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol


class Encoder(Protocol):
    def __call__(self, obj: Any) -> Any:
        ...



from dataclasses import dataclass

from xpresso.binders._parameters.openapi.base import (
    OpenAPIParameterBase,
    OpenAPIParameterMarkerBase,
)
from xpresso.openapi import models as openapi_models


class OpenAPIPathParameter(OpenAPIParameterBase):
    param_cls = openapi_models.Path


@dataclass(frozen=True)
class OpenAPIPathParameterMarker(OpenAPIParameterMarkerBase):
    cls = OpenAPIPathParameter
    in_ = "path"
    required = True



import sys
import typing

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

import starlette.middleware
from starlette.routing import BaseRoute
from starlette.routing import Router as StarletteRouter
from starlette.types import Receive, Scope, Send

from xpresso.dependencies.models import Dependant
from xpresso.responses import Responses


class _ASGIApp(Protocol):
    def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> typing.Awaitable[None]:
        ...


_MiddlewareIterator = typing.Iterable[
    typing.Tuple[typing.Callable[..., _ASGIApp], typing.Mapping[str, typing.Any]]
]


class Router:
    routes: typing.Sequence[BaseRoute]
    lifespan: typing.Optional[typing.Callable[..., typing.AsyncContextManager[None]]]
    dependencies: typing.Sequence[Dependant]
    tags: typing.Sequence[str]
    responses: Responses
    include_in_schema: bool
    _app: _ASGIApp

    __slots__ = (
        "_app",
        "_router",
        "dependencies",
        "include_in_schema",
        "lifespan",
        "responses",
        "routes",
        "tags",
    )

    def __init__(
        self,
        routes: typing.Sequence[BaseRoute],
        *,
        middleware: typing.Optional[
            typing.Sequence[starlette.middleware.Middleware]
        ] = None,
        lifespan: typing.Optional[
            typing.Callable[..., typing.AsyncContextManager[None]]
        ] = None,
        redirect_slashes: bool = True,
        default: typing.Optional[_ASGIApp] = None,
        dependencies: typing.Optional[typing.Sequence[Dependant]] = None,
        tags: typing.Optional[typing.List[str]] = None,
        responses: typing.Optional[Responses] = None,
        include_in_schema: bool = True,
    ) -> None:
        self.routes = list(routes)
        self.lifespan = lifespan
        self._router = StarletteRouter(
            routes=self.routes,
            redirect_slashes=redirect_slashes,
            default=default,  # type: ignore[arg-type]
            lifespan=lifespan,  # type: ignore[arg-type]
        )
        self.dependencies = list(dependencies or [])
        self.tags = list(tags or [])
        self.responses = dict(responses or {})
        self.include_in_schema = include_in_schema
        self._app = self._router.__call__
        if middleware is not None:
            for cls, options in typing.cast(_MiddlewareIterator, reversed(middleware)):
                self._app = cls(app=self._app, **options)

    def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> typing.Awaitable[None]:
        return self._app(scope, receive, send)



import inspect
import typing
from dataclasses import dataclass

from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import ModelField
from starlette.datastructures import FormData, UploadFile
from starlette.requests import Request

from xpresso._utils.media_type_validator import MediaTypeValidator
from xpresso._utils.media_type_validator import (
    get_validator as get_media_type_validator,
)
from xpresso._utils.typing import model_field_from_param
from xpresso.binders._body.extractors.body_field_validation import validate_body_field
from xpresso.binders._utils.stream_to_bytes import convert_stream_to_bytes
from xpresso.binders.api import BodyExtractor, BodyExtractorMarker
from xpresso.exceptions import RequestValidationError
from xpresso.typing import Some


@dataclass(frozen=True)
class FileBodyExtractor(BodyExtractor):
    field: ModelField
    media_type_validator: MediaTypeValidator
    consume: bool

    def matches_media_type(self, media_type: typing.Optional[str]) -> bool:
        return self.media_type_validator.matches(media_type)

    async def extract_from_request(self, request: Request) -> typing.Any:
        media_type = request.headers.get("content-type", None)
        self.media_type_validator.validate(media_type, loc=("body",))
        if self.field.type_ is bytes:
            if self.consume:
                data = await convert_stream_to_bytes(request.stream())
                if data is None:
                    return validate_body_field(
                        Some(b""), field=self.field, loc=("body",)
                    )
            else:
                data = await request.body()
            return validate_body_field(Some(data), field=self.field, loc=("body",))
        # create an UploadFile from the body's stream
        file: UploadFile = self.field.type_(  # use the field type to allow users to subclass UploadFile
            filename="body", content_type=media_type or "*/*"
        )
        non_empty_chunks = 0
        chunks = 0
        async for chunk in request.stream():
            non_empty_chunks += len(chunk) != 0
            chunks += 1
            await file.write(chunk)
        await file.seek(0)
        return file

    async def extract_from_field(
        self,
        field: typing.Union[str, UploadFile],
        *,
        loc: typing.Iterable[typing.Union[str, int]],
    ) -> typing.Any:
        return await self._extract(field, loc=loc)

    async def extract_from_form(
        self, form: FormData, *, loc: typing.Iterable[typing.Union[str, int]]
    ) -> typing.Optional[Some[typing.Any]]:
        raise NotImplementedError

    async def _extract(
        self,
        value: typing.Union[str, UploadFile],
        loc: typing.Iterable[typing.Union[int, str]],
    ) -> typing.Union[bytes, UploadFile]:
        if isinstance(value, str):
            raise RequestValidationError(
                [
                    ErrorWrapper(
                        exc=TypeError("Expected a file, got a string"),
                        loc=tuple(loc),
                    )
                ]
            )
        if self.field.type_ is bytes:
            # user requested bytes
            return await value.read()  # type: ignore  # UploadFile always returns bytes
        return value

    # These are implemented to work around pecularities of hashing bound methods on Python 3.7
    def __hash__(self) -> int:
        return id(self)  # pragma: no covet

    def __eq__(self, __o: object) -> bool:
        return self is __o  # pragma: no covet


@dataclass(frozen=True)
class FileBodyExtractorMarker(BodyExtractorMarker):
    media_type: typing.Optional[str]
    enforce_media_type: bool
    consume: bool

    def register_parameter(self, param: inspect.Parameter) -> BodyExtractor:
        field = model_field_from_param(param)
        if self.media_type and self.enforce_media_type:
            media_type_validator = get_media_type_validator(self.media_type)
        else:
            media_type_validator = get_media_type_validator(None)
        return FileBodyExtractor(
            field=field,
            media_type_validator=media_type_validator,
            consume=self.consume,
        )



from typing import Any, Dict, Iterable, List, Mapping, Optional, Set, Tuple, cast

from di import BaseContainer
from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY

from xpresso._utils.routing import VisitedRoute
from xpresso._utils.typing import get_model_name_map
from xpresso.binders import dependants as binder_dependants
from xpresso.openapi import models
from xpresso.openapi._responses import (
    get_response,
    get_response_specs_from_return_type_hints,
    merge_response_models,
)
from xpresso.openapi._security import get_security_models, get_security_scheme_name_map
from xpresso.openapi.constants import REF_PREFIX
from xpresso.responses import Responses, ResponseSpec
from xpresso.routing.operation import Operation
from xpresso.routing.pathitem import Path
from xpresso.routing.router import Router

ModelNameMap = Dict[type, str]

Routes = Mapping[str, Tuple[Path, Mapping[str, Operation]]]

SecurityModels = Mapping[binder_dependants.SecurityBinder, models.SecurityScheme]

validation_error_definition = {
    "title": "ValidationError",
    "type": "object",
    "properties": {
        "loc": {
            "title": "Location",
            "type": "array",
            "items": {"oneOf": [{"type": "string"}, {"type": "integer"}]},
        },
        "msg": {"title": "Message", "type": "string"},
        "type": {"title": "Error Type", "type": "string"},
    },
    "required": ["loc", "msg", "type"],
}

validation_error_response_definition = {
    "title": "HTTPValidationError",
    "type": "object",
    "properties": {
        "detail": {
            "title": "Detail",
            "type": "array",
            "items": {"$ref": REF_PREFIX + "ValidationError"},
        }
    },
}


def get_parameters(
    deps: List[binder_dependants.ParameterBinder],
    model_name_map: ModelNameMap,
    schemas: Dict[str, Any],
) -> Optional[List[models.ConcreteParameter]]:
    parameters: List[models.ConcreteParameter] = []
    for dependant in deps:
        if dependant.openapi and dependant.openapi.include_in_schema:
            parameters.append(
                dependant.openapi.get_openapi(
                    model_name_map=model_name_map, schemas=schemas
                )
            )
    if parameters:
        return list(sorted(parameters, key=lambda param: param.name))
    return None


def get_request_body(
    dependant: binder_dependants.BodyBinder,
    model_name_map: ModelNameMap,
    schemas: Dict[str, Any],
) -> models.RequestBody:
    if dependant.openapi and dependant.openapi.include_in_schema:
        return dependant.openapi.get_openapi(
            model_name_map=model_name_map, schemas=schemas
        )
    return models.RequestBody(content={})


def get_responses(
    route: Operation,
    response_specs: Responses,
    model_name_map: ModelNameMap,
    schemas: Dict[str, Any],
) -> Dict[str, models.Response]:
    responses: Dict[str, models.Response] = {}
    for status_code, response in response_specs.items():
        status = str(status_code)
        if (
            status in responses
            or f"{status[0]}XX" in responses
            or (
                status.endswith("XX")
                and any(s.startswith(status[0]) for s in responses)
            )
        ):
            raise ValueError("Duplicate response status codes are not allowed")
        if isinstance(response, ResponseSpec):
            model = get_response(response, model_name_map, schemas)
        else:
            # iterable of response specs
            model = merge_response_models(
                (get_response(r, model_name_map, schemas) for r in response),
                default_description="Successful Response",
            )
        responses[status] = model
    if responses:
        return responses
    responses_from_type_hints = get_response_specs_from_return_type_hints(
        route.endpoint
    )
    responses_from_type_hints = responses_from_type_hints or [
        ResponseSpec(description="Successful Response")
    ]
    response_model = merge_response_models(
        (
            get_response(spec, model_name_map, schemas)
            for spec in responses_from_type_hints
        ),
        default_description="Successful Response",
    )
    return {"200": response_model}


def get_operation(
    route: Operation,
    model_name_map: ModelNameMap,
    components: Dict[str, Any],
    security_models: Mapping[binder_dependants.SecurityBinder, str],
    tags: List[str],
    response_specs: Responses,
) -> models.Operation:
    data: Dict[str, Any] = {
        "tags": tags or None,
        "summary": route.summary,
        "description": route.description,
        "deprecated": route.deprecated,
        "servers": route.servers or None,
        "external_docs": route.external_docs,
    }
    docstring = getattr(route.endpoint, "__doc__", None)
    if docstring and not data["description"]:
        data["description"] = docstring
    schemas: Dict[str, Any] = {}
    route_dependant = route.dependant
    assert route_dependant is not None
    parameters = get_parameters(
        [
            dep
            for dep in route_dependant.get_flat_subdependants()
            if isinstance(dep, binder_dependants.ParameterBinder)
        ],
        model_name_map,
        schemas,
    )
    if parameters:
        data["parameters"] = parameters
    body_dependant = next(
        (
            dep
            for dep in route_dependant.get_flat_subdependants()
            if isinstance(dep, binder_dependants.BodyBinder)
        ),
        None,
    )
    if body_dependant is not None:
        data["requestBody"] = get_request_body(body_dependant, model_name_map, schemas)
    security_dependants: List[binder_dependants.SecurityBinder] = [
        dep
        for dep in route_dependant.get_flat_subdependants()
        if isinstance(dep, binder_dependants.SecurityBinder)
    ]
    if security_dependants:
        security_schemes = components.get("securitySchemes", None) or {}
        components["securitySchemes"] = security_schemes
        data["security"] = [
            {security_models[dep]: list(dep.scopes)}
            for dep in sorted(security_dependants, key=lambda dep: security_models[dep])
        ]
    data["responses"] = get_responses(
        route,
        response_specs=response_specs,
        model_name_map=model_name_map,
        schemas=schemas,
    )
    if not data["responses"]:
        data["responses"] = {"200": {"description": "Successful Response"}}
    if schemas:
        components["schemas"] = {**components.get("schemas", {}), **schemas}
    http422 = str(HTTP_422_UNPROCESSABLE_ENTITY)
    if (data.get("parameters", None) or data.get("requestBody", None)) and not any(
        status in data["responses"] for status in (http422, "4XX", "default")
    ):
        data["responses"][http422] = {
            "description": "Validation Error",
            "content": {
                "application/json": {
                    "schema": {"$ref": REF_PREFIX + "HTTPValidationError"}
                }
            },
        }
        if "ValidationError" not in schemas:
            components["schemas"] = components.get("schemas", None) or {}
            components["schemas"].update(
                {
                    "ValidationError": validation_error_definition,
                    "HTTPValidationError": validation_error_response_definition,
                }
            )
    return models.Operation(**data)


def get_paths_items(
    visitor: Iterable[VisitedRoute[Any]],
    model_name_map: ModelNameMap,
    components: Dict[str, Any],
    security_models: Mapping[binder_dependants.SecurityBinder, str],
) -> Dict[str, models.PathItem]:
    paths: Dict[str, models.PathItem] = {}
    for visited_route in visitor:
        if isinstance(visited_route.route, Path):
            path_item = visited_route.route
            if not path_item.include_in_schema:
                continue
            tags: List[str] = []
            responses = dict(cast(Responses, {}))
            include_in_schema = True
            for node in visited_route.nodes:
                if isinstance(node, Router):
                    if not node.include_in_schema:
                        include_in_schema = False
                        break
                    responses.update(node.responses)
                    tags.extend(node.tags)
            if not include_in_schema:
                continue
            tags.extend(path_item.tags)
            responses.update(path_item.responses)
            operations: Dict[str, models.Operation] = {}
            for method, operation in path_item.operations.items():
                if not operation.include_in_schema:
                    continue
                operations[method.lower()] = get_operation(
                    operation,
                    model_name_map=model_name_map,
                    components=components,
                    security_models=security_models,
                    tags=tags + operation.tags,
                    response_specs={**responses, **operation.responses},
                )
            paths[visited_route.path] = models.PathItem(
                description=visited_route.route.description,
                summary=visited_route.route.summary,
                servers=visited_route.route.servers or None,
                **operations,  # type: ignore[arg-type]
            )  # type: ignore  # for Pylance
    return {k: paths[k] for k in sorted(paths.keys())}


def filter_routes(visitor: Iterable[VisitedRoute[Any]]) -> Routes:
    res: Dict[str, Tuple[Path, Dict[str, Operation]]] = {}
    for visited_route in visitor:
        if isinstance(visited_route.route, Path):
            path_item = visited_route.route
            if not path_item.include_in_schema:
                continue
            operations: Dict[str, Operation] = {}
            for method, operation in path_item.operations.items():
                if not operation.include_in_schema:
                    continue
                operations[method.lower()] = operation
            res[visited_route.path] = (path_item, operations)
    return res


def get_flat_models(routes: Routes) -> Set[type]:
    res: Set[type] = set()
    for _, operations in routes.values():
        for operation in operations.values():
            dependant = operation.dependant
            flat_dependencies = dependant.get_flat_subdependants()
            for dep in flat_dependencies:
                if isinstance(
                    dep,
                    (binder_dependants.ParameterBinder, binder_dependants.BodyBinder),
                ):
                    openapi = dep.openapi
                    res.update(openapi.get_models())
    return res


def genrate_openapi(
    visitor: Iterable[VisitedRoute[Any]],
    container: BaseContainer,
    version: str,
    info: models.Info,
    servers: Optional[Iterable[models.Server]],
) -> models.OpenAPI:
    visitor = list(visitor)
    routes = filter_routes(visitor)
    flat_models = get_flat_models(routes)
    model_name_map = get_model_name_map(flat_models)
    security_binders = get_security_models(routes, container)
    security_scheme_name_map = get_security_scheme_name_map(security_binders)
    security_bider_to_name_map = {
        binder: security_scheme_name_map[scheme]
        for binder, scheme in security_binders.items()
    }
    components_security_schemes = {
        name: scheme for scheme, name in security_scheme_name_map.items()
    }
    components: Dict[str, Any] = {}
    if components_security_schemes:
        components["securitySchemes"] = {
            name: components_security_schemes[name]
            for name in sorted(components_security_schemes.keys())
        }
    paths = get_paths_items(
        visitor, model_name_map, components, security_bider_to_name_map
    )
    return models.OpenAPI(
        openapi=version,
        info=info,
        paths=paths,  # type: ignore[arg-type]
        components=models.Components(**components) if components else None,
        servers=list(servers) if servers else None,
    )



import functools
import inspect
import sys
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, Iterable, List, Optional, Tuple, cast

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import ModelField
from starlette.requests import HTTPConnection

from xpresso._utils.typing import is_mapping_like, is_sequence_like
from xpresso.binders._parameters.extractors.base import (
    ParameterExtractorBase,
    get_basic_param_info,
)
from xpresso.binders._utils.grouped import grouped
from xpresso.binders.api import ParameterExtractor, ParameterExtractorMarker
from xpresso.binders.exceptions import InvalidSerialization
from xpresso.exceptions import RequestValidationError, WebSocketValidationError
from xpresso.typing import Some


def collect_scalar(value: Optional[str]) -> Optional[Some[str]]:
    if value is None:
        return None
    split = value.split(",")
    if len(split) == 1:
        return Some(split[0])
    return Some(next(iter(split)))


def collect_sequence(value: Optional[str]) -> Optional[Some[List[str]]]:
    if not value:
        return Some(cast(List[str], []))
    return Some([v.lstrip() for v in value.split(",")])


def collect_object(
    explode: bool,
    value: Optional[str],
) -> Optional[Some[Dict[str, str]]]:
    if not value:
        return None
    if explode:
        res: Dict[str, str] = {}
        for field in value.split(","):
            split = field.split("=", maxsplit=1)
            if len(split) == 1 or not field or field[0] == "=":
                raise InvalidSerialization(f"invalid object style header: {value}")
            name, val = split
            res[name.lstrip()] = val
        return Some(res)
    try:
        groups = cast(
            Iterable[Tuple[str, str]], grouped([v.lstrip() for v in value.split(",")])
        )
    except ValueError:
        raise InvalidSerialization(f"invalid object style header: {value}")
    return Some(dict(groups))


class Extractor(Protocol):
    def __call__(self, value: Optional[str]) -> Optional[Some[Any]]:
        ...


def get_extractor(explode: bool, field: ModelField) -> Extractor:
    # form style
    if is_sequence_like(field):
        return collect_sequence
    if is_mapping_like(field):
        return functools.partial(collect_object, explode)
    # single item
    return collect_scalar


ERRORS = {
    "webscoket": WebSocketValidationError,
    "http": RequestValidationError,
}


@dataclass(frozen=True)
class HeaderParameterExtractor(ParameterExtractorBase):
    extractor: Extractor
    in_: ClassVar[str] = "header"
    header_name: bytes

    async def extract(
        self,
        connection: HTTPConnection,
    ) -> Any:
        # parse headers according to RFC 7230
        # this means treating repeated headers and "," seperated ones the same
        # so here we merge them all into one "," seperated string
        # also note that whitespaces after a "," don't matter, so we .lstrip() as needed
        header_values: "List[str]" = []
        for name, value in connection.scope["headers"]:
            if name == self.header_name:
                header_values.append(value.decode("latin-1"))
        header_value: "Optional[str]"
        if header_values:
            header_value = ",".join(header_values)
        else:
            header_value = None
        try:
            extracted = self.extractor(header_value)
        except InvalidSerialization as exc:
            raise ERRORS[connection.scope["type"]](
                [ErrorWrapper(exc=exc, loc=("header", self.name))]
            )
        return await self.validate(extracted, connection)


@dataclass(frozen=True)
class HeaderParameterExtractorMarker(ParameterExtractorMarker):
    alias: Optional[str]
    explode: bool
    convert_underscores: bool
    in_: ClassVar[str] = "header"

    def register_parameter(self, param: inspect.Parameter) -> ParameterExtractor:
        field, name, loc = get_basic_param_info(param, self.alias, self.in_)
        if self.convert_underscores and not self.alias and field.name == field.alias:
            name = name.replace("_", "-")
        extractor = get_extractor(explode=self.explode, field=field)
        return HeaderParameterExtractor(
            field=field,
            loc=loc,
            name=name,
            extractor=extractor,
            header_name=name.lower().encode("latin-1"),
        )



from __future__ import annotations

import sys
import typing

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

from di import BaseContainer
from starlette.background import BackgroundTasks
from starlette.requests import HTTPConnection, Request
from starlette.websockets import WebSocket

from xpresso.dependencies.models import Dependant


class App(Protocol):
    @property
    def container(self) -> BaseContainer:
        ...


def register_framework_dependencies(
    container: BaseContainer, app: App, app_type: typing.Type[App]
):
    container.register_by_type(
        Dependant(Request, scope="connection", wire=False),
        Request,
    )
    container.register_by_type(
        Dependant(Request, scope="connection", wire=False),
        Request,
    )
    container.register_by_type(
        Dependant(
            HTTPConnection,
            scope="connection",
            wire=False,
        ),
        HTTPConnection,
    )
    container.register_by_type(
        Dependant(
            WebSocket,
            scope="connection",
            wire=False,
        ),
        WebSocket,
    )
    container.register_by_type(
        Dependant(
            lambda: BackgroundTasks(),
            scope="connection",
            wire=False,
        ),
        BackgroundTasks,
    )
    container.register_by_type(
        Dependant(
            lambda: app.container,
            scope="app",
            wire=False,
        ),
        BaseContainer,
        covariant=True,
    )
    container.register_by_type(
        Dependant(
            lambda: app,
            scope="app",
            wire=False,
        ),
        app_type,
        covariant=True,
    )



import inspect
import sys
from typing import Dict, List, Optional, Union

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

from xpresso.openapi import models

Model = type
ModelNameMap = Dict[Model, str]
Schemas = Dict[str, Union[models.Schema, models.Reference]]


class OpenAPIBody(Protocol):
    include_in_schema: bool

    def get_models(self) -> List[type]:
        """The type representing this parameter.

        This is used as the key to deduplicate references and schema names.
        """
        raise NotImplementedError

    def get_openapi(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.RequestBody:
        raise NotImplementedError

    def get_schema(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.Schema:
        raise NotImplementedError

    def get_media_type_string(self) -> str:
        raise NotImplementedError

    def get_openapi_media_type(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.MediaType:
        raise NotImplementedError

    def get_encoding(self) -> Optional[models.Encoding]:
        raise NotImplementedError

    def get_field_name(self) -> str:
        raise NotImplementedError


class OpenAPIBodyMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> OpenAPIBody:
        raise NotImplementedError


class OpenAPIParameter(Protocol):
    in_: str
    name: str
    include_in_schema: bool

    def get_openapi(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.ConcreteParameter:
        raise NotImplementedError

    def get_models(self) -> List[type]:
        """Used as the key to deduplicate references and schema names"""
        raise NotImplementedError


class OpenAPIParameterMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> OpenAPIParameter:
        raise NotImplementedError


class OpenAPISecurityScheme(Protocol):
    def get_openapi(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.SecurityScheme:
        raise NotImplementedError

    def get_models(self) -> List[type]:
        """Used as the key to deduplicate references and schema names"""
        raise NotImplementedError


class OpenAPISecuritySchemeMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> OpenAPISecurityScheme:
        raise NotImplementedError
